import argparse
import numpy as np
import magicmind.python.runtime as mm
import cv2
from calibrator import CalibData
from magicmind.python.runtime.parser import Parser

def torch_parser(args):
    # 创建MagicMind parser
    parser = Parser(mm.ModelKind.kPytorch)
    # 设置网络输入数据类型
    parser.set_model_param("pytorch-input-dtypes", [mm.DataType.FLOAT32])
    # 创建一个空的网络实例
    network = mm.Network()
    # 使用parser将PyTorch模型文件转换为MagicMind Network实例。
    assert parser.parse(network, args.pt_model).ok()
    # 设置模型输入形状
    input_dims = mm.Dims((args.batch_size, 1, args.input_height, args.input_width))
    assert network.get_input(0).set_dimension(input_dims).ok()
    return network

def generate_model_config(args):
    config = mm.BuilderConfig()
    # 指定硬件平台
    assert config.parse_from_string('{"archs":[{"mtp_372": [2,6,8]}]}').ok()
    # INT64 转 INT32
    assert config.parse_from_string('{"opt_config":{"type64to32_conversion":true}}').ok()
    assert config.parse_from_string('{"opt_config":{"conv_scale_fold":true}}').ok()
    # 输入数据摆放顺序
    # PyTorch模型输入数据顺序为NCHW，如下代码转为NHWC输入顺序。
    # 输入顺序的改变需要同步到推理过程中的网络预处理实现，保证预处理结果的输入顺序与网络输入数据顺序一致。
    # 以下JSON字符串中的0代表改变的是网络第一个输入的数据摆放顺序。1则代表第二个输入，以此类推。
    assert config.parse_from_string('{"convert_input_layout": {"0": {"src": "NCHW", "dst": "NHWC"}}}').ok()
    assert config.parse_from_string('{"convert_output_layout": {"0": {"src": "NCHW", "dst": "NHWC"}}}').ok()
    # 模型输入输出规模可变功能
    if args.shape_mutable == "true":
        assert config.parse_from_string('{"graph_shape_mutable":true}').ok()
        assert config.parse_from_string('{"dim_range": {"0": {"min": [1, 1, 320, 256], "max": [64, 1, 320, 256]}}}').ok()
    else:
        assert config.parse_from_string('{"graph_shape_mutable":false}').ok()
    # 精度模式
    assert config.parse_from_string('{"precision_config":{"precision_mode":"%s"}}' % args.precision).ok()
    # 量化算法，支持对称量化（symmetric)和非对称量化（asymmetric）。当量化统计算法设置为EQNM_ALOGORITHM时，仅适用于对称量化。
    assert config.parse_from_string('{"precision_config": {"activation_quant_algo": "symmetric"}}').ok()
    # 设置量化粒度，支持按tensor量化（per_tensor）和按通道量化（per_axis）两种。
    assert config.parse_from_string('{"precision_config": {"weight_quant_granularity": "per_axis"}}').ok()
    return config

def calibrate(args, network : mm.Network, config : mm.BuilderConfig):
    # 创建量化工具并设置量化统计算法
    calib_data = CalibData(args)
    calibrator = mm.Calibrator([calib_data])
    assert calibrator is not None
    # 设置量化统计算法，支持线性统计算法（LINEAR_ALGORITHM）及加强的最小化量化噪声算法（EQM_ALGORITHM）。
    assert calibrator.set_quantization_algorithm(mm.QuantizationAlgorithm.LINEAR_ALGORITHM).ok()
    # 打开设备
    with mm.System() as mm_sys:
        dev_count = mm_sys.device_count()
        print("Device count: ", dev_count)
        if args.device_id >= dev_count:
            print("Invalid device set!")
            abort()
        # 打开MLU设备
        dev = mm.Device()
        dev.id = args.device_id
        assert dev.active().ok()
    # 进行量化
    assert calibrator.calibrate(network, config).ok()

def main():
    args = argparse.ArgumentParser()
    args.add_argument('--pt_model', dest = 'pt_model', default = '/workspace/saved_pt/2d_unet_0.pt',
            required = True, type = str, help = 'tf output graph')
    args.add_argument('--output_model', dest = 'output_model', default = '/workspace/offline_models/2d_unet_0.model',
            type = str, help = 'output model path')
    args.add_argument('--shape_mutable', dest = 'shape_mutable', default = 'false',
            type = str, help = 'whether the mm model is dynamic or static or not"')
    args.add_argument('--precision', dest = 'precision', default = 'qint8_mixed_float16',
            type = str, help = 'precision, qint8_mixed_float16 qint8_mixed_float32 force_float16 force float32 are supported')
    args.add_argument('--calib_data_path', dest = 'calib_data_path', default = 'pytorch_2d_unet_calib_data.pt',
            type = str, help = 'image list file path, file contains input image paths for calibration')
    args.add_argument('--batch_size', dest = 'batch_size', default = 4,
            type = int, help = 'batch_size')
    args.add_argument('--input_width', dest = 'input_width', default = 256,
            type = int, help = 'model input width')
    args.add_argument('--input_height', dest = 'input_height', default = 320,
            type = int, help = 'model input height')
    args.add_argument('--device_id', dest = 'device_id', default = 0,
            type = int, help = 'mlu device id, used for calibration')
    args = args.parse_args()

    network = torch_parser(args)
    config = generate_model_config(args)
    if args.precision.find('qint') != -1:
        print('do calibrate...')
        calibrate(args, network, config)
    print('build model...')
    # 生成模型
    builder = mm.Builder()
    model = builder.build_model('magicmind model', network, config)
    assert model is not None
    # 将模型序列化为离线文件
    assert model.serialize_to_file(args.output_model).ok()
    print("Generate model done, model save to %s" % args.output_model)

if __name__ == "__main__":
    main()



import magicmind.python.runtime as mm
from magicmind.python.common.types import get_datatype_by_numpy
import torch
import numpy as np
class CalibData(mm.CalibDataInterface):
        def __init__(self, args):
            super().__init__()
            self.count = 1
            self.pth_path = args.calib_data_path
            self.batch_size=args.batch_size

        def get_shape(self):
            return mm.Dims(self.data[0].shape)

        def get_data_type(self):
            return get_datatype_by_numpy(self.data[0].dtype)

        def get_sample(self):
            return self.data[0]

        def next(self):
            if self.count:
                self.data = [np.repeat(torch.load(self.pth_path).numpy(), self.batch_size, axis=0)]
                self.count = self.count - 1
            else:
                return mm.Status(mm.Code.OUT_OF_RANGE, "No more data.")
            return mm.Status.OK()

        def reset(self):
            self.count = 1
            return mm.Status.OK()



#!/usr/bin/python3

import os
import csv
import argparse
import sys

root_dir = os.getenv("PROJ_ROOT_PATH")
csv_path = os.path.join(root_dir, "benchmark", "benchmark.csv")
fieldnames = ["model", "language", "dataset","metric", "eval", "fps"]

parser = argparse.ArgumentParser()
for key in fieldnames:
    parser.add_argument("--%s"%key)

def write_result(**kw):
    model = kw.get("model", str(os.environ.get("PROJ_ROOT_PATH")).split("/")[-1])
    value = {k:kw.get(k,None) for k in fieldnames[1:]}
    value = {k: v for k, v in value.items() if v is not None}

    if os.path.exists(csv_path):
        with open(csv_path, "r") as f:
            reader = csv.DictReader(f)
            rows = [x for x in reader]
    else:
        rows = []
    for row in rows:
        if model  == row[fieldnames[0]]:
            row.update(value)
    if model not in [x[fieldnames[0]] for x in rows]:
        row = value
        row[fieldnames[0]] = model
        rows.append(row)
    with open(csv_path,'w',newline='')as c:
        writer = csv.DictWriter(c, fieldnames=fieldnames)
        if len(rows) == 1:
            writer.writeheader()
        writer.writerows(rows)

if __name__ == "__main__":
    param = parser.parse_args()
    dic = vars(param)
    dic = {k: v for k, v in dic.items() if v is not None}
    write_result(**dic)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2014-2016  Brno University of Technology (author: Karel Vesely)
# Licensed under the Apache License, Version 2.0 (the "License")

import numpy as np
import sys, os, re, gzip, struct

#################################################
# Adding kaldi tools to shell path,

# Select kaldi,
if not 'KALDI_ROOT' in os.environ:
  # Default! To change run python with 'export KALDI_ROOT=/some_dir python'
  os.environ['KALDI_ROOT']='/mnt/matylda5/iveselyk/Tools/kaldi-trunk'

# Add kaldi tools to path,
os.environ['PATH'] = os.popen('echo $KALDI_ROOT/src/bin:$KALDI_ROOT/tools/openfst/bin:$KALDI_ROOT/src/fstbin/:$KALDI_ROOT/src/gmmbin/:$KALDI_ROOT/src/featbin/:$KALDI_ROOT/src/lm/:$KALDI_ROOT/src/sgmmbin/:$KALDI_ROOT/src/sgmm2bin/:$KALDI_ROOT/src/fgmmbin/:$KALDI_ROOT/src/latbin/:$KALDI_ROOT/src/nnetbin:$KALDI_ROOT/src/nnet2bin:$KALDI_ROOT/src/nnet3bin:$KALDI_ROOT/src/online2bin/:$KALDI_ROOT/src/ivectorbin/:$KALDI_ROOT/src/lmbin/').readline().strip() + ':' + os.environ['PATH']


#################################################
# Define all custom exceptions,
class UnsupportedDataType(Exception): pass
class UnknownVectorHeader(Exception): pass
class UnknownMatrixHeader(Exception): pass

class BadSampleSize(Exception): pass
class BadInputFormat(Exception): pass

class SubprocessFailed(Exception): pass

#################################################
# Data-type independent helper functions,

def open_or_fd(file, mode='rb'):
  """ fd = open_or_fd(file)
   Open file, gzipped file, pipe, or forward the file-descriptor.
   Eventually seeks in the 'file' argument contains ':offset' suffix.
  """
  offset = None
  try:
    # strip 'ark:' prefix from r{x,w}filename (optional),
    if re.search('^(ark|scp)(,scp|,b|,t|,n?f|,n?p|,b?o|,n?s|,n?cs)*:', file):
      (prefix,file) = file.split(':',1)
    # separate offset from filename (optional),
    if re.search(':[0-9]+$', file):
      (file,offset) = file.rsplit(':',1)
    # input pipe?
    if file[-1] == '|':
      fd = popen(file[:-1], 'rb') # custom,
    # output pipe?
    elif file[0] == '|':
      fd = popen(file[1:], 'wb') # custom,
    # is it gzipped?
    elif file.split('.')[-1] == 'gz':
      fd = gzip.open(file, mode)
    # a normal file...
    else:
      fd = open(file, mode)
  except TypeError:
    # 'file' is opened file descriptor,
    fd = file
  # Eventually seek to offset,
  if offset != None: fd.seek(int(offset))
  return fd

# based on '/usr/local/lib/python3.4/os.py'
def popen(cmd, mode="rb"):
  if not isinstance(cmd, str):
    raise TypeError("invalid cmd type (%s, expected string)" % type(cmd))

  import subprocess, io, threading

  # cleanup function for subprocesses,
  def cleanup(proc, cmd):
    ret = proc.wait()
    if ret > 0:
      raise SubprocessFailed('cmd %s returned %d !' % (cmd,ret))
    return

  # text-mode,
  if mode == "r":
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    threading.Thread(target=cleanup,args=(proc,cmd)).start() # clean-up thread,
    return io.TextIOWrapper(proc.stdout)
  elif mode == "w":
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
    threading.Thread(target=cleanup,args=(proc,cmd)).start() # clean-up thread,
    return io.TextIOWrapper(proc.stdin)
  # binary,
  elif mode == "rb":
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    threading.Thread(target=cleanup,args=(proc,cmd)).start() # clean-up thread,
    return proc.stdout
  elif mode == "wb":
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
    threading.Thread(target=cleanup,args=(proc,cmd)).start() # clean-up thread,
    return proc.stdin
  # sanity,
  else:
    raise ValueError("invalid mode %s" % mode)


def read_key(fd):
  """ [key] = read_key(fd)
   Read the utterance-key from the opened ark/stream descriptor 'fd'.
  """
  key = ''
  while 1:
    char = fd.read(1).decode("latin1")
    if char == '' : break
    if char == ' ' : break
    key += char
  key = key.strip()
  if key == '': return None # end of file,
  assert(re.match('^\S+$',key) != None) # check format (no whitespace!)
  return key


#################################################
# Integer vectors (alignments, ...),

def read_ali_ark(file_or_fd):
  """ Alias to 'read_vec_int_ark()' """
  return read_vec_int_ark(file_or_fd)

def read_vec_int_ark(file_or_fd):
  """ generator(key,vec) = read_vec_int_ark(file_or_fd)
   Create generator of (key,vector<int>) tuples, which reads from the ark file/stream.
   file_or_fd : ark, gzipped ark, pipe or opened file descriptor.

   Read ark to a 'dictionary':
   d = { u:d for u,d in kaldi_io.read_vec_int_ark(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    key = read_key(fd)
    while key:
      ali = read_vec_int(fd)
      yield key, ali
      key = read_key(fd)
  finally:
    if fd is not file_or_fd: fd.close()

def read_vec_int_scp(file_or_fd):
  """ generator(key,vec) = read_vec_int_scp(file_or_fd)
   Returns generator of (key,vector<int>) tuples, read according to kaldi scp.
   file_or_fd : scp, gzipped scp, pipe or opened file descriptor.

   Iterate the scp:
   for key,vec in kaldi_io.read_vec_int_scp(file):
     ...

   Read scp to a 'dictionary':
   d = { key:vec for key,mat in kaldi_io.read_vec_int_scp(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    for line in fd:
      (key,rxfile) = line.decode().split(' ')
      vec = read_vec_int(rxfile)
      yield key, vec
  finally:
    if fd is not file_or_fd : fd.close()

def read_vec_int(file_or_fd):
  """ [int-vec] = read_vec_int(file_or_fd)
   Read kaldi integer vector, ascii or binary input,
  """
  fd = open_or_fd(file_or_fd)
  binary = fd.read(2).decode()
  if binary == '\0B': # binary flag
    assert(fd.read(1).decode() == '\4'); # int-size
    vec_size = np.frombuffer(fd.read(4), dtype='int32', count=1)[0] # vector dim
    # Elements from int32 vector are sored in tuples: (sizeof(int32), value),
    vec = np.frombuffer(fd.read(vec_size*5), dtype=[('size','int8'),('value','int32')], count=vec_size)
    assert(vec[0]['size'] == 4) # int32 size,
    ans = vec[:]['value'] # values are in 2nd column,
  else: # ascii,
    arr = (binary + fd.readline().decode()).strip().split()
    try:
      arr.remove('['); arr.remove(']') # optionally
    except ValueError:
      pass
    ans = np.array(arr, dtype=int)
  if fd is not file_or_fd : fd.close() # cleanup
  return ans

# Writing,
def write_vec_int(file_or_fd, v, key=''):
  """ write_vec_int(f, v, key='')
   Write a binary kaldi integer vector to filename or stream.
   Arguments:
   file_or_fd : filename or opened file descriptor for writing,
   v : the vector to be stored,
   key (optional) : used for writing ark-file, the utterance-id gets written before the vector.

   Example of writing single vector:
   kaldi_io.write_vec_int(filename, vec)

   Example of writing arkfile:
   with open(ark_file,'w') as f:
     for key,vec in dict.iteritems():
       kaldi_io.write_vec_flt(f, vec, key=key)
  """
  fd = open_or_fd(file_or_fd, mode='wb')
  if sys.version_info[0] == 3: assert(fd.mode == 'wb')
  try:
    if key != '' : fd.write((key+' ').encode("latin1")) # ark-files have keys (utterance-id),
    fd.write('\0B'.encode()) # we write binary!
    # dim,
    fd.write('\4'.encode()) # int32 type,
    fd.write(struct.pack(np.dtype('int32').char, v.shape[0]))
    # data,
    for i in range(len(v)):
      fd.write('\4'.encode()) # int32 type,
      fd.write(struct.pack(np.dtype('int32').char, v[i])) # binary,
  finally:
    if fd is not file_or_fd : fd.close()


#################################################
# Float vectors (confidences, ivectors, ...),

# Reading,
def read_vec_flt_scp(file_or_fd):
  """ generator(key,mat) = read_vec_flt_scp(file_or_fd)
   Returns generator of (key,vector) tuples, read according to kaldi scp.
   file_or_fd : scp, gzipped scp, pipe or opened file descriptor.

   Iterate the scp:
   for key,vec in kaldi_io.read_vec_flt_scp(file):
     ...

   Read scp to a 'dictionary':
   d = { key:mat for key,mat in kaldi_io.read_mat_scp(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    for line in fd:
      (key,rxfile) = line.decode().split(' ')
      vec = read_vec_flt(rxfile)
      yield key, vec
  finally:
    if fd is not file_or_fd : fd.close()

def read_vec_flt_ark(file_or_fd):
  """ generator(key,vec) = read_vec_flt_ark(file_or_fd)
   Create generator of (key,vector<float>) tuples, reading from an ark file/stream.
   file_or_fd : ark, gzipped ark, pipe or opened file descriptor.

   Read ark to a 'dictionary':
   d = { u:d for u,d in kaldi_io.read_vec_flt_ark(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    key = read_key(fd)
    while key:
      ali = read_vec_flt(fd)
      yield key, ali
      key = read_key(fd)
  finally:
    if fd is not file_or_fd: fd.close()

def read_vec_flt(file_or_fd):
  """ [flt-vec] = read_vec_flt(file_or_fd)
   Read kaldi float vector, ascii or binary input,
  """
  fd = open_or_fd(file_or_fd)
  binary = fd.read(2).decode()
  if binary == '\0B': # binary flag
    # Data type,
    header = fd.read(3).decode()
    if header == 'FV ': sample_size = 4 # floats
    elif header == 'DV ': sample_size = 8 # doubles
    else: raise UnknownVectorHeader("The header contained '%s'" % header)
    assert(sample_size > 0)
    # Dimension,
    assert(fd.read(1).decode() == '\4'); # int-size
    vec_size = np.frombuffer(fd.read(4), dtype='int32', count=1)[0] # vector dim
    # Read whole vector,
    buf = fd.read(vec_size * sample_size)
    if sample_size == 4 : ans = np.frombuffer(buf, dtype='float32')
    elif sample_size == 8 : ans = np.frombuffer(buf, dtype='float64')
    else : raise BadSampleSize
    return ans
  else: # ascii,
    arr = (binary + fd.readline().decode()).strip().split()
    try:
      arr.remove('['); arr.remove(']') # optionally
    except ValueError:
      pass
    ans = np.array(arr, dtype=float)
  if fd is not file_or_fd : fd.close() # cleanup
  return ans

# Writing,
def write_vec_flt(file_or_fd, v, key=''):
  """ write_vec_flt(f, v, key='')
   Write a binary kaldi vector to filename or stream. Supports 32bit and 64bit floats.
   Arguments:
   file_or_fd : filename or opened file descriptor for writing,
   v : the vector to be stored,
   key (optional) : used for writing ark-file, the utterance-id gets written before the vector.

   Example of writing single vector:
   kaldi_io.write_vec_flt(filename, vec)

   Example of writing arkfile:
   with open(ark_file,'w') as f:
     for key,vec in dict.iteritems():
       kaldi_io.write_vec_flt(f, vec, key=key)
  """
  fd = open_or_fd(file_or_fd, mode='wb')
  if sys.version_info[0] == 3: assert(fd.mode == 'wb')
  try:
    if key != '' : fd.write((key+' ').encode("latin1")) # ark-files have keys (utterance-id),
    fd.write('\0B'.encode()) # we write binary!
    # Data-type,
    if v.dtype == 'float32': fd.write('FV '.encode())
    elif v.dtype == 'float64': fd.write('DV '.encode())
    else: raise UnsupportedDataType("'%s', please use 'float32' or 'float64'" % v.dtype)
    # Dim,
    fd.write('\04'.encode())
    fd.write(struct.pack(np.dtype('uint32').char, v.shape[0])) # dim
    # Data,
    fd.write(v.tobytes())
  finally:
    if fd is not file_or_fd : fd.close()


#################################################
# Float matrices (features, transformations, ...),

# Reading,
def read_mat_scp(file_or_fd):
  """ generator(key,mat) = read_mat_scp(file_or_fd)
   Returns generator of (key,matrix) tuples, read according to kaldi scp.
   file_or_fd : scp, gzipped scp, pipe or opened file descriptor.

   Iterate the scp:
   for key,mat in kaldi_io.read_mat_scp(file):
     ...

   Read scp to a 'dictionary':
   d = { key:mat for key,mat in kaldi_io.read_mat_scp(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    for line in fd:
      (key,rxfile) = line.decode().split(' ')
      mat = read_mat(rxfile)
      yield key, mat
  finally:
    if fd is not file_or_fd : fd.close()

def read_mat_ark(file_or_fd):
  """ generator(key,mat) = read_mat_ark(file_or_fd)
   Returns generator of (key,matrix) tuples, read from ark file/stream.
   file_or_fd : scp, gzipped scp, pipe or opened file descriptor.

   Iterate the ark:
   for key,mat in kaldi_io.read_mat_ark(file):
     ...

   Read ark to a 'dictionary':
   d = { key:mat for key,mat in kaldi_io.read_mat_ark(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    key = read_key(fd)
    while key:
      mat = read_mat(fd)
      yield key, mat
      key = read_key(fd)
  finally:
    if fd is not file_or_fd : fd.close()

def read_mat(file_or_fd):
  """ [mat] = read_mat(file_or_fd)
   Reads single kaldi matrix, supports ascii and binary.
   file_or_fd : file, gzipped file, pipe or opened file descriptor.
  """
  fd = open_or_fd(file_or_fd)
  try:
    binary = fd.read(2).decode()
    if binary == '\0B' :
      mat = _read_mat_binary(fd)
    else:
      assert(binary == ' [')
      mat = _read_mat_ascii(fd)
  finally:
    if fd is not file_or_fd: fd.close()
  return mat

def _read_mat_binary(fd):
  # Data type
  header = fd.read(3).decode()
  # 'CM', 'CM2', 'CM3' are possible values,
  if header.startswith('CM'): return _read_compressed_mat(fd, header)
  elif header == 'FM ': sample_size = 4 # floats
  elif header == 'DM ': sample_size = 8 # doubles
  else: raise UnknownMatrixHeader("The header contained '%s'" % header)
  assert(sample_size > 0)
  # Dimensions
  s1, rows, s2, cols = np.frombuffer(fd.read(10), dtype='int8,int32,int8,int32', count=1)[0]
  # Read whole matrix
  buf = fd.read(rows * cols * sample_size)
  if sample_size == 4 : vec = np.frombuffer(buf, dtype='float32')
  elif sample_size == 8 : vec = np.frombuffer(buf, dtype='float64')
  else : raise BadSampleSize
  mat = np.reshape(vec,(rows,cols))
  return mat

def _read_mat_ascii(fd):
  rows = []
  while 1:
    line = fd.readline().decode()
    if (len(line) == 0) : raise BadInputFormat # eof, should not happen!
    if len(line.strip()) == 0 : continue # skip empty line
    arr = line.strip().split()
    if arr[-1] != ']':
      rows.append(np.array(arr,dtype='float32')) # not last line
    else:
      rows.append(np.array(arr[:-1],dtype='float32')) # last line
      mat = np.vstack(rows)
      return mat


def _read_compressed_mat(fd, format):
  """ Read a compressed matrix,
      see: https://github.com/kaldi-asr/kaldi/blob/master/src/matrix/compressed-matrix.h
      methods: CompressedMatrix::Read(...), CompressedMatrix::CopyToMat(...),
  """
  assert(format == 'CM ') # The formats CM2, CM3 are not supported...

  # Format of header 'struct',
  global_header = np.dtype([('minvalue','float32'),('range','float32'),('num_rows','int32'),('num_cols','int32')]) # member '.format' is not written,
  per_col_header = np.dtype([('percentile_0','uint16'),('percentile_25','uint16'),('percentile_75','uint16'),('percentile_100','uint16')])

  # Mapping for percentiles in col-headers,
  def uint16_to_float(value, min, range):
    return np.float32(min + range * 1.52590218966964e-05 * value)

  # Mapping for matrix elements,
  def uint8_to_float_v2(vec, p0, p25, p75, p100):
    # Split the vector by masks,
    mask_0_64 = (vec <= 64);
    mask_193_255 = (vec > 192);
    mask_65_192 = (~(mask_0_64 | mask_193_255));
    # Sanity check (useful but slow...),
    # assert(len(vec) == np.sum(np.hstack([mask_0_64,mask_65_192,mask_193_255])))
    # assert(len(vec) == np.sum(np.any([mask_0_64,mask_65_192,mask_193_255], axis=0)))
    # Build the float vector,
    ans = np.empty(len(vec), dtype='float32')
    ans[mask_0_64] = p0 + (p25 - p0) / 64. * vec[mask_0_64]
    ans[mask_65_192] = p25 + (p75 - p25) / 128. * (vec[mask_65_192] - 64)
    ans[mask_193_255] = p75 + (p100 - p75) / 63. * (vec[mask_193_255] - 192)
    return ans

  # Read global header,
  globmin, globrange, rows, cols = np.frombuffer(fd.read(16), dtype=global_header, count=1)[0]

  # The data is structed as [Colheader, ... , Colheader, Data, Data , .... ]
  #                         {           cols           }{     size         }
  col_headers = np.frombuffer(fd.read(cols*8), dtype=per_col_header, count=cols)
  data = np.reshape(np.frombuffer(fd.read(cols*rows), dtype='uint8', count=cols*rows), newshape=(cols,rows)) # stored as col-major,

  mat = np.empty((cols,rows), dtype='float32')
  for i, col_header in enumerate(col_headers):
    col_header_flt = [ uint16_to_float(percentile, globmin, globrange) for percentile in col_header ]
    mat[i] = uint8_to_float_v2(data[i], *col_header_flt)

  return mat.T # transpose! col-major -> row-major,

def write_ark_scp(key, mat, ark_fout, scp_out):
  mat_offset = write_mat(ark_fout, mat, key)
  scp_line = '{}\t{}:{}'.format(key, ark_fout.name, mat_offset)
  scp_out.write(scp_line)
  scp_out.write('\n')

# Writing,
def write_mat(file_or_fd, m, key=''):
  """ write_mat(f, m, key='')
  Write a binary kaldi matrix to filename or stream. Supports 32bit and 64bit floats.
  Arguments:
   file_or_fd : filename of opened file descriptor for writing,
   m : the matrix to be stored,
   key (optional) : used for writing ark-file, the utterance-id gets written before the matrix.

   Example of writing single matrix:
   kaldi_io.write_mat(filename, mat)

   Example of writing arkfile:
   with open(ark_file,'w') as f:
     for key,mat in dict.iteritems():
       kaldi_io.write_mat(f, mat, key=key)
  """
  mat_offset = 0
  fd = open_or_fd(file_or_fd, mode='wb')
  if sys.version_info[0] == 3: assert(fd.mode == 'wb')
  try:
    if key != '' : fd.write((key+' ').encode("latin1")) # ark-files have keys (utterance-id),
    mat_offset = fd.tell()
    fd.write('\0B'.encode()) # we write binary!
    # Data-type,
    if m.dtype == 'float32': fd.write('FM '.encode())
    elif m.dtype == 'float64': fd.write('DM '.encode())
    else: raise UnsupportedDataType("'%s', please use 'float32' or 'float64'" % m.dtype)
    # Dims,
    fd.write('\04'.encode())
    fd.write(struct.pack(np.dtype('uint32').char, m.shape[0])) # rows
    fd.write('\04'.encode())
    fd.write(struct.pack(np.dtype('uint32').char, m.shape[1])) # cols
    # Data,
    fd.write(m.tobytes())
  finally:
    if fd is not file_or_fd : fd.close()
  return mat_offset

#################################################
# 'Posterior' kaldi type (posteriors, confusion network, nnet1 training targets, ...)
# Corresponds to: vector<vector<tuple<int,float> > >
# - outer vector: time axis
# - inner vector: records at the time
# - tuple: int = index, float = value
#

def read_cnet_ark(file_or_fd):
  """ Alias of function 'read_post_ark()', 'cnet' = confusion network """
  return read_post_ark(file_or_fd)

def read_post_ark(file_or_fd):
  """ generator(key,vec<vec<int,float>>) = read_post_ark(file)
   Returns generator of (key,posterior) tuples, read from ark file.
   file_or_fd : ark, gzipped ark, pipe or opened file descriptor.

   Iterate the ark:
   for key,post in kaldi_io.read_post_ark(file):
     ...

   Read ark to a 'dictionary':
   d = { key:post for key,post in kaldi_io.read_post_ark(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    key = read_key(fd)
    while key:
      post = read_post(fd)
      yield key, post
      key = read_key(fd)
  finally:
    if fd is not file_or_fd: fd.close()

def read_post(file_or_fd):
  """ [post] = read_post(file_or_fd)
   Reads single kaldi 'Posterior' in binary format.

   The 'Posterior' is C++ type 'vector<vector<tuple<int,float> > >',
   the outer-vector is usually time axis, inner-vector are the records
   at given time,  and the tuple is composed of an 'index' (integer)
   and a 'float-value'. The 'float-value' can represent a probability
   or any other numeric value.

   Returns vector of vectors of tuples.
  """
  fd = open_or_fd(file_or_fd)
  ans=[]
  binary = fd.read(2).decode(); assert(binary == '\0B'); # binary flag
  assert(fd.read(1).decode() == '\4'); # int-size
  outer_vec_size = np.frombuffer(fd.read(4), dtype='int32', count=1)[0] # number of frames (or bins)

  # Loop over 'outer-vector',
  for i in range(outer_vec_size):
    assert(fd.read(1).decode() == '\4'); # int-size
    inner_vec_size = np.frombuffer(fd.read(4), dtype='int32', count=1)[0] # number of records for frame (or bin)
    data = np.frombuffer(fd.read(inner_vec_size*10), dtype=[('size_idx','int8'),('idx','int32'),('size_post','int8'),('post','float32')], count=inner_vec_size)
    assert(data[0]['size_idx'] == 4)
    assert(data[0]['size_post'] == 4)
    ans.append(data[['idx','post']].tolist())

  if fd is not file_or_fd: fd.close()
  return ans


#################################################
# Kaldi Confusion Network bin begin/end times,
# (kaldi stores CNs time info separately from the Posterior).
#

def read_cntime_ark(file_or_fd):
  """ generator(key,vec<tuple<float,float>>) = read_cntime_ark(file_or_fd)
   Returns generator of (key,cntime) tuples, read from ark file.
   file_or_fd : file, gzipped file, pipe or opened file descriptor.

   Iterate the ark:
   for key,time in kaldi_io.read_cntime_ark(file):
     ...

   Read ark to a 'dictionary':
   d = { key:time for key,time in kaldi_io.read_post_ark(file) }
  """
  fd = open_or_fd(file_or_fd)
  try:
    key = read_key(fd)
    while key:
      cntime = read_cntime(fd)
      yield key, cntime
      key = read_key(fd)
  finally:
    if fd is not file_or_fd : fd.close()

def read_cntime(file_or_fd):
  """ [cntime] = read_cntime(file_or_fd)
   Reads single kaldi 'Confusion Network time info', in binary format:
   C++ type: vector<tuple<float,float> >.
   (begin/end times of bins at the confusion network).

   Binary layout is '<num-bins> <beg1> <end1> <beg2> <end2> ...'

   file_or_fd : file, gzipped file, pipe or opened file descriptor.

   Returns vector of tuples.
  """
  fd = open_or_fd(file_or_fd)
  binary = fd.read(2).decode(); assert(binary == '\0B'); # assuming it's binary

  assert(fd.read(1).decode() == '\4'); # int-size
  vec_size = np.frombuffer(fd.read(4), dtype='int32', count=1)[0] # number of frames (or bins)

  data = np.frombuffer(fd.read(vec_size*10), dtype=[('size_beg','int8'),('t_beg','float32'),('size_end','int8'),('t_end','float32')], count=vec_size)
  assert(data[0]['size_beg'] == 4)
  assert(data[0]['size_end'] == 4)
  ans = data[['t_beg','t_end']].tolist() # Return vector of tuples (t_beg,t_end),

  if fd is not file_or_fd : fd.close()
  return ans


#################################################
# Segments related,
#

# Segments as 'Bool vectors' can be handy,
# - for 'superposing' the segmentations,
# - for frame-selection in Speaker-ID experiments,
def read_segments_as_bool_vec(segments_file):
  """ [ bool_vec ] = read_segments_as_bool_vec(segments_file)
   using kaldi 'segments' file for 1 wav, format : '<utt> <rec> <t-beg> <t-end>'
   - t-beg, t-end is in seconds,
   - assumed 100 frames/second,
  """
  segs = np.loadtxt(segments_file, dtype='object,object,f,f', ndmin=1)
  # Sanity checks,
  assert(len(segs) > 0) # empty segmentation is an error,
  assert(len(np.unique([rec[1] for rec in segs ])) == 1) # segments with only 1 wav-file,
  # Convert time to frame-indexes,
  start = np.rint([100 * rec[2] for rec in segs]).astype(int)
  end = np.rint([100 * rec[3] for rec in segs]).astype(int)
  # Taken from 'read_lab_to_bool_vec', htk.py,
  frms = np.repeat(np.r_[np.tile([False,True], len(end)), False],
                   np.r_[np.c_[start - np.r_[0, end[:-1]], end-start].flat, 0])
  assert np.sum(end-start) == np.sum(frms)
  return frms




import os
import json
import numpy as np
import cv2
import argparse
import logging
import magicmind.python.runtime as mm
import magicmind.python.runtime.parser
from PIL import Image 
from calibrator import CalibData, Calibrator

parser = argparse.ArgumentParser()
parser.add_argument("--tf_model", type=str, default= "../data/models/frozen_inference_graph.pb", help="tf frozen pb model")
parser.add_argument("--image_dir",  type=str, default=os.path.join(str(os.environ.get("VOC2012_DATASETS_PATH")), "VOC2012","JPEGImages"), help="VOC2012 datasets")
parser.add_argument("--file_list", type=str, default=os.path.join(str(os.environ.get("VOC2012_DATASETS_PATH")), "VOC2012","ImageSets","Segmentation","val.txt"), help="val.txt path")
parser.add_argument("--output_model_path", type=str, default= "../data/models/deeplabv3_tensorflow_model_force_float32_true_1", help="save mm model to this path")
parser.add_argument("--precision", type=str, default="force_float32", help="qint8_mixed_float16, qint8_mixed_float32, qint16_mixed_float16, qint16_mixed_float32, force_float32, force_float16")
def voc_dataset(file_list, image_file_path, count):
    with open(file_list, "r") as f:
        lines = f.readlines()
    logging.info("%d pictures will be read." % len(lines))
    current_count = 0
    for line in lines:
        image_name = line.replace("\n", "")
        image_path = os.path.join(image_file_path, image_name + ".jpg")
        img = Image.open(image_path)
        yield img
        current_count += 1
        if current_count > count and count != -1:
            break

if __name__ == "__main__":
    args = parser.parse_args()
    input_size = 513
    batch_size = 1
    # 创建 MagicMind Network
    deeplabv3_network = mm.Network()
     # 创建 MagicMind Config
    config = mm.BuilderConfig()
    # 创建 MagicMind Parser （tensorflow后端）
    tf_parser = mm.parser.Parser(mm.ModelKind.kTensorflow)
    tf_parser.set_model_param("tf-model-type", "tf-graphdef-file")
    tf_parser.set_model_param("tf-graphdef-inputs",["ImageTensor:0"])
    tf_parser.set_model_param("tf-graphdef-outputs", ["SemanticPredictions:0"])
    tf_parser.set_model_param("tf-infer-shape", True)
    assert tf_parser.parse(deeplabv3_network, args.tf_model).ok()

    deeplabv3_network.get_input(0).set_dimension(mm.Dims([batch_size, input_size, input_size, 3]))
    # 指定硬件平台
    assert config.parse_from_string('{"archs":[{"mtp_372": [2,6,8]}]}').ok()
    # INT64 转 INT32
    assert config.parse_from_string('{"opt_config":{"type64to32_conversion":true}}').ok()
    # conv_scale_fold
    assert config.parse_from_string('{"opt_config":{"conv_scale_fold":true}}').ok()
    # 模型输入输出规模可变功能
    assert config.parse_from_string('{"graph_shape_mutable":false}').ok()
    # 精度模式
    assert config.parse_from_string('{"precision_config":{"precision_mode":"%s"}}' % args.precision).ok()

    dataset = voc_dataset(file_list = args.file_list, image_file_path = args.image_dir, count = batch_size)
    sample_data = []
    for data in dataset:
        width, height = data.size
        resize_ratio = 1.0 * input_size / max(width, height)
        target_size = (input_size, input_size)
        resized_image = data.convert('RGB').resize(target_size, Image.ANTIALIAS)
        resized_image = np.asarray(resized_image)
        input_data = np.copy(resized_image)
        sample_data.append(np.expand_dims(input_data, 0))

    if "qint" in args.precision:
        calib_data = CalibData(sample_data)
        calibrator = Calibrator(calib_data, mm.QuantizationAlgorithm.LINEAR_ALGORITHM)
        assert config.parse_from_string('{"precision_config": {"weight_quant_granularity": "per_tensor"}}').ok()
        assert calibrator.calibrate(deeplabv3_network, config).ok()

    builder = mm.Builder()
    model = builder.build_model("deeplabv3_mm_model", deeplabv3_network, config)
    assert model != None
    model.serialize_to_file(args.output_model_path)
    logging.info("Generate model done, model save to %s" % args.output_model_path) 



import magicmind.python.runtime as mm
import argparse
import numpy as np
import torch
import os
import sys
import torch.nn.functional as F
from tqdm import tqdm
from torch import Tensor
from PIL import Image
from data_loading import BasicDataset, CarvanaDataset
from torch.utils.data import DataLoader, random_split

from mm_runner import MMRunner
from logger import Logger

log = Logger()
IMAGE_HEIGHT = 640
IMAGE_WIDTH = 959



MODEL_PATH = os.getenv("MODEL_PATH")
CARVANA_DATASETS_PATH = os.getenv("CARVANA_DATASETS_PATH")
PROJ_ROOT_PATH = os.getenv("PROJ_ROOT_PATH")

parser = argparse.ArgumentParser()
parser.add_argument(
        "--device_id", dest="device_id", type=int, default=0, help="device_id"
)
parser.add_argument(
        "--magicmind_model", dest="magicmind_model",type=str, default=str(MODEL_PATH) + "/unet_carvana_scale0.5_epoch2.mm", help="magicmind_model"
)
parser.add_argument(
        "--data_folder",  dest="data_folder",type=str, default = str(CARVANA_DATASETS_PATH)
)
parser.add_argument(
        "--output_folder",  dest="output_folder",type=str, default=  "../data/output",
                    help="forder for saving output."
)
parser.add_argument(
        "--classes", "-c",  dest="classes",type=int, default=2, help="Number of classes"
)
parser.add_argument(
        "--scale", "-s",  dest="scale",type=float, default=0.5, help="Downscaling factor of the images"
)
parser.add_argument(
        "--batch_size", "-b", dest="batch_size", metavar="B", type=int, default=1, help="Batch size"
)
parser.add_argument(
        "--validation", "-v", dest="val", type=float, default=10.0,
                        help="Percent of the data that is used as validation (0-100)"
)
parser.add_argument(
        "--save_img",  dest="save_img",action="store_true", default=False , help="Save the output masks"
)
parser.add_argument(
        "--mask-threshold","-t", dest="mask-threshold", type=float, default=0.5,
                        help='Minimum probability value to consider a mask pixel white'
)
parser.add_argument(
    "--image_num", dest="image_num", type=int, default=10, help="image number"
)


def dice_coeff(input: Tensor, target: Tensor, reduce_batch_first: bool = False, epsilon: float = 1e-6):
    # Average of Dice coefficient for all batches, or for a single mask
    assert input.size() == target.size()
    assert input.dim() == 3 or not reduce_batch_first

    sum_dim = (-1, -2) if input.dim() == 2 or not reduce_batch_first else (-1, -2, -3)

    inter = 2 * (input * target).sum(dim=sum_dim)
    sets_sum = input.sum(dim=sum_dim) + target.sum(dim=sum_dim)
    sets_sum = torch.where(sets_sum == 0, inter, sets_sum)

    dice = (inter + epsilon) / (sets_sum + epsilon)
    return dice.mean()

def multiclass_dice_coeff(input: Tensor, target: Tensor, reduce_batch_first: bool = False, epsilon: float = 1e-6):
    # Average of Dice coefficient for all classes
    return dice_coeff(input.flatten(0, 1), target.flatten(0, 1), reduce_batch_first, epsilon)


def mask_to_image(mask: np.ndarray, mask_values):
    if isinstance(mask_values[0], list):
        out = np.zeros((mask.shape[-2], mask.shape[-1], len(mask_values[0])), dtype=np.uint8)
    elif mask_values == [0, 1]:
        out = np.zeros((mask.shape[-2], mask.shape[-1]), dtype=bool)
    else:
        out = np.zeros((mask.shape[-2], mask.shape[-1]), dtype=np.uint8)

    if mask.ndim == 3:
        mask = np.argmax(mask, axis=0)

    for i, v in enumerate(mask_values):
        out[mask == i] = v

    return Image.fromarray(out)

class MM_Model(object):
    def __init__(self, model_path,device_id):
        #mlu
        self.model = MMRunner(mm_file=model_path, device_id=device_id)
    def forward(self, input):
        outputs = []
        inputs = [input]
        outputs = self.model(inputs)

        return outputs[0]

def main():
    args = parser.parse_args()
    if not os.path.exists(args.magicmind_model):
        print("please generate magicmind model first!!!")
        exit()

    Unet = MM_Model(args.magicmind_model,args.device_id)
    dir_img = args.data_folder + '/imgs/'
    dir_mask = args.data_folder + '/masks/'
    # 1. Create dataset
    try:
        dataset = CarvanaDataset(dir_img, dir_mask, args.scale)
    except (AssertionError, RuntimeError, IndexError):
        dataset = BasicDataset(dir_img, dir_mask, args.scale)

    # 2. Split into train / validation partitions
    n_val = int(len(dataset) * args.val / 100)
    n_train = len(dataset) - n_val
    train_set, val_set = random_split(dataset, [n_train, n_val], generator=torch.Generator().manual_seed(0))

    # 3. Create data loaders
    loader_args = dict(batch_size=args.batch_size, num_workers= 0, pin_memory=True)
    val_loader = DataLoader(val_set, shuffle=False, drop_last=False, **loader_args)

    num_val_batches = len(val_loader)
    dice_score = 0
    number = 0
    count = 0
    log.info("Start run ...")
    image_num = args.image_num
    #image_num = num_val_batches
    batch_size=args.batch_size
    if image_num >= 508:
        image_num = 508
    rem_img_num = image_num % batch_size
    img_idx = 0
    batch_counter = 0
    total_tmp = int(image_num/batch_size)
    imgs = np.empty([batch_size, 3, IMAGE_HEIGHT, IMAGE_WIDTH])
    if total_tmp*batch_size < image_num:
        total_tmp += 1
    for batch in tqdm(val_loader, total=total_tmp, desc='Validation round', unit='batch', leave=False):
        image_ori, mask_true = batch['image'], batch['mask']
        infer_batch_size = (
            batch_size if img_idx < (image_num - rem_img_num) else rem_img_num
        )
        if infer_batch_size == 0:
            break
        # predict the mask
        #image = image_ori.permute(0,2,3,1)
        mask_trues = torch.split(mask_true,1,dim=0)
        for index in range (0,infer_batch_size):
            image = image_ori[index]
            imgs[batch_counter % infer_batch_size, :, :, :] = image
            batch_counter += 1
            img_idx += 1
        if batch_counter % infer_batch_size == 0:
            batch_counter = 0
            output = Unet.forward(imgs)
            mask_pred = torch.tensor(output.astype(np.float32))
            mask_preds = torch.split(mask_pred,1,dim=0)
            for pred_idx in range(infer_batch_size):
                if args.save_img:
                    out_filename = str(number) + '.jpg'
                    mask_values = [0, 1]
                    pad_h = int(image_ori.shape[3] / args.scale)
                    pad_w = int(image_ori.shape[2] / args.scale)
                    output = F.interpolate(mask_pred, (pad_w, pad_h), mode='bilinear')
                    if args.classes > 1:
                        mask = output.argmax(dim=1)
                    else:
                        mask = torch.sigmoid(output) > args.mask_threshold
                    make_return = mask[0].long().squeeze().numpy()
                    result = mask_to_image(make_return, mask_values)
                    result.save(args.output_folder + os.sep + out_filename)
                    number = number + 1
                    print('Mask saved to ' + args.output_folder + os.sep + out_filename)

                if args.classes == 1:
                    assert mask_true.min() >= 0 and mask_true.max() <= 1, 'True mask indices should be in [0, 1]'
                    mask_pred = (F.sigmoid(mask_pred) > 0.5).float()
                    # compute the Dice score
                    dice_score += dice_coeff(mask_pred, mask_true, reduce_batch_first=False)
                else:
                    assert mask_true.min() >= 0 and mask_true.max() < args.classes, 'True mask indices should be in [0, classes['
                    # convert to one-hot format
                    mask_true1 = F.one_hot(mask_trues[pred_idx], args.classes).permute(0, 3, 1, 2).float()
                    mask_pred1 = F.one_hot(mask_preds[pred_idx].argmax(dim=1), args.classes).permute(0, 3, 1, 2).float()
                    # compute the Dice score, ignoring background
                    dice_score += multiclass_dice_coeff(mask_pred1[:, 1:], mask_true1[:, 1:], reduce_batch_first=False)
                count += 1
    val_score =  dice_score / max(image_num, 1)
    print("Dice coefficient:",val_score)
    f = open(args.output_folder + os.sep + "result.txt", "w")
    f.write('Dice coefficient: %.4f\n' %(val_score))
    f.close()


if __name__ == "__main__":
    main()



#! -*- coding: utf-8 -*-
import argparse
from bert4keras.tokenizers import Tokenizer
from bert4keras.snippets import sequence_padding
import magicmind.python.runtime as mm
from mm_runner import MMRunner
from logger import Logger
logging = Logger()
maxlen = 64

def similarity(text1, text2, tokenizer, model):
    """"计算text1与text2的相似度
    """
    texts = [text1, text2]
    X, S = [], []
    for t in texts:
        x, s = tokenizer.encode(t, maxlen=maxlen)
        X.append(x)
        S.append(s)
    X = sequence_padding(X)
    S = sequence_padding(S)
    Z = model([S, X])[0]
    Z /= (Z**2).sum(axis=1, keepdims=True)**0.5
    return (Z[0] * Z[1]).sum()


def main():
    args = argparse.ArgumentParser()
    args.add_argument("--magicmind_model", type=str, default="../data/models/roformer-sim_tf_force_float32_false_2_64", help="mm model")
    args.add_argument("--vocab_file", type=str, default="../data/models/chinese_roformer-sim-char-ft_L-6_H-384_A-6/vocab.txt", help="dict path")
    args.add_argument("--device_id", type=int, default=0, help="which device use")
    args = args.parse_args()

    tokenizer = Tokenizer(args.vocab_file, do_lower_case=True)  # 建立分词器 
    model = MMRunner(mm_file=args.magicmind_model, device_id=args.device_id)  

    strings_1 = [u'今天天气不错', u'今天天气不错',u'我喜欢北京',u'我喜欢北京']
    strings_2 = [u'今天天气很好', u'今天天气不好', u'我很喜欢北京', u'我不喜欢北京']
    for i in range(len(strings_1)):
        smi = similarity(strings_1[i], strings_2[i], tokenizer, model)
        logging.info('{}, {}, similarity is : {}'.format(strings_1[i], strings_2[i], smi))

if __name__ == "__main__":
    main()


import magicmind.python.runtime as mm
import numpy as np
import cv2
from typing import List
from logger import Logger

log = Logger()


class CalibData(mm.CalibDataInterface):
    def __init__(
        self, shape: mm.Dims, max_samples: int, resize_h, resize_w, video_list: List
    ):
        super().__init__()
        self.captures_ = []
        for path in video_list:
            self.captures_.append(cv2.VideoCapture(path.strip()))
        self.shape_ = shape
        self.max_samples_ = max_samples
        self.cur_capture_index_ = 0
        self.processed_clips_ = 0
        self.resize_w = resize_w
        self.resize_h = resize_h

    def get_shape(self):
        return self.shape_

    def get_data_type(self):
        return mm.DataType.FLOAT32

    def get_sample(self):
        return self.cur_sample_

    # read a clip
    def read_clip(self):
        clip_len = self.shape_.GetDimValue(2)
        while self.cur_capture_index_ < len(self.captures_):
            capture = self.captures_[self.cur_capture_index_]
            clip = []
            while len(clip) < clip_len:
                got_frame, frame = capture.read()
                if not got_frame:
                    break
                else:
                    clip.append(frame)
            if len(clip) < clip_len:
                # reach the end of the video, read next video
                self.cur_capture_index_ = self.cur_capture_index_ + 1
            else:
                return clip
        log.info("get None")
        # no more data
        return None

    def next(self):
        if self.processed_clips_ >= self.max_samples_:
            return mm.Status(mm.Code.OUT_OF_RANGE, "Maximum samples reached")
        batch_size = self.shape_.GetDimValue(0)
        clip_len = self.shape_.GetDimValue(2)
        h = self.shape_.GetDimValue(3)
        w = self.shape_.GetDimValue(4)
        clips = []
        for i in range(batch_size):
            clip = self.read_clip()
            if clip is None:
                # no more data
                return mm.Status(mm.Code.OUT_OF_RANGE, "Data end reached")
            clips.append(clip)

        # resize and center crop
        preprocessed_clips = []
        for clip in clips:
            preprocessed_clip = []
            for frame in clip:
                resized = cv2.resize(frame, (self.resize_w, self.resize_h))
                top = round((self.resize_h - h) / 2)
                left = round((self.resize_w - w) / 2)
                cropped = resized[top : top + h, left : left + w, :]
                normalized = cropped - [128, 128, 128]
                preprocessed_clip.append(normalized)
            preprocessed_clip = np.array(preprocessed_clip)  # DHWC
            preprocessed_clip = preprocessed_clip.transpose((3, 0, 1, 2))  # CDHW
            preprocessed_clips.append(preprocessed_clip)
        self.cur_sample_ = np.array(preprocessed_clips).astype("float32")  # NCDHW
        self.processed_clips_ = self.processed_clips_ + 1
        return mm.Status.OK()

    def reset(self):
        self.cur_sample_ = None
        self.cur_capture_index_ = 0
        self.processed_clips_ = 0
        return mm.Status.OK()



import magicmind.python.runtime as mm
import argparse
import numpy as np
import cv2
import os
import glob
import math
import logging
from PIL import Image

def letterbox(img, dst_shape):
    src_h, src_w = img.shape[0], img.shape[1]
    dst_h, dst_w = dst_shape
    ratio = min(dst_h / src_h, dst_w / src_w)
    unpad_h, unpad_w = int(math.floor(src_h * ratio)), int(math.floor(src_w * ratio))
    if ratio != 1:
        interp = cv2.INTER_AREA if ratio < 1 else cv2.INTER_LINEAR
        img = cv2.resize(img, (unpad_w, unpad_h), interpolation=interp)
    # padding
    pad_t = int(math.floor((dst_h - unpad_h) / 2))
    pad_b = dst_h - unpad_h - pad_t
    pad_l = int(math.floor((dst_w - unpad_w) / 2))
    pad_r = dst_w - unpad_w - pad_l
    img = cv2.copyMakeBorder(img, pad_t, pad_b, pad_l, pad_r, cv2.BORDER_CONSTANT, value=(114,114,114))
    return img, ratio

def preprocess_image(img, dst_shape) -> np.ndarray:
    # resize as letterbox
    img, ratio = letterbox(img, dst_shape)
    # BGR to RGB, HWC to CHW
    img = img[:, :, ::-1].transpose(2, 0, 1)
    # normalize
    img = img.astype(dtype = np.float32) / 255.0
    return img


class CalibData(mm.CalibDataInterface):
    def __init__(self, shape: mm.Dims, max_samples: int, img_dir: str):
        super().__init__()
        # print(img_dir)
        assert os.path.isdir(img_dir)
        self.data_paths_ = glob.glob(img_dir + '/*.jpg')
        self.shape_ = shape
        self.max_samples_ = min(max_samples, len(self.data_paths_))
        self.cur_sample_ = None
        self.cur_data_index_ = 0
        self.dst_shape_ = (self.shape_.GetDimValue(2), self.shape_.GetDimValue(3))

    def get_shape(self):
        return self.shape_

    def get_data_type(self):
        return mm.DataType.FLOAT32

    def get_sample(self):
        return self.cur_sample_

    def preprocess_images(self, data_begin: int, data_end: int) -> np.ndarray:
        imgs = []
        for i in range(data_begin, data_end):
            img = cv2.imread(self.data_paths_[i])
            img = preprocess_image(img, self.dst_shape_)
            imgs.append(img[np.newaxis,:])
        # batch and normalize
        return np.ascontiguousarray(np.concatenate(tuple(imgs), axis=0))

    def next(self):
        batch_size = self.shape_.GetDimValue(0)
        data_begin = self.cur_data_index_
        data_end = data_begin + batch_size
        if data_end > self.max_samples_:
            return mm.Status(mm.Code.OUT_OF_RANGE, "Data end reached")
        self.cur_sample_ =  self.preprocess_images(data_begin, data_end)
        self.cur_data_index_ = data_end
        return mm.Status.OK()

    def reset(self):
        self.cur_sample_ = None
        self.cur_data_index_ = 0
        return mm.Status.OK()



import argparse
import os
import numpy as np
from skimage import transform, io
import magicmind.python.runtime as mm
from magicmind.python.runtime import DataType
from magicmind.python.common.types import get_numpy_dtype_by_datatype
import cv2
from magicmind.python.runtime.parser import Parser

MSRA_B_DATASETS_PATH = os.getenv("MSRA_B_DATASETS_PATH")

def torch_parser(args):
    # 创建MagicMind parser
    parser = Parser(mm.ModelKind.kPytorch)
    # 设置网络输入数据类型
    parser.set_model_param("pytorch-input-dtypes", [mm.DataType.FLOAT32])
    # 创建一个空的网络实例
    network = mm.Network()
    # 使用parser将PyTorch模型文件转换为MagicMind Network实例。
    assert parser.parse(network, args.pt_model).ok()
    # 设置模型输入形状
    input_dims = mm.Dims((args.batch_size, 3, args.input_height, args.input_width))
    assert network.get_input(0).set_dimension(input_dims).ok()
    return network

def generate_model_config(args):
    config = mm.BuilderConfig()
    # 指定硬件平台
    assert config.parse_from_string('{"archs":[{"mtp_372": [6,8]}]}').ok()
    assert config.parse_from_string('{"opt_config":{"conv_scale_fold":true}}').ok()
    assert config.parse_from_string('{"precision_config":{"precision_mode":"%s"}}' % args.precision).ok()
    return config

def preprocess(img):
    output_size = (320, 320)
    new_h, new_w = output_size
    image = transform.resize(img,(new_h, new_w),mode='constant')
    tmpImg = np.zeros((image.shape[0], image.shape[1],3))
    image = image/np.max(image)
    if image.shape[2] == 1:
        tmpImg[:,:,0] = (image[:,:,0]-0.485)/0.229
        tmpImg[:,:,1] = (image[:,:,0]-0.485)/0.229
        tmpImg[:,:,2] = (image[:,:,0]-0.485)/0.229
    else:
        tmpImg[:,:,0] = (image[:,:,0]-0.485)/0.229
        tmpImg[:,:,1] = (image[:,:,1]-0.456)/0.224
        tmpImg[:,:,2] = (image[:,:,2]-0.406)/0.225
    tmpImg = tmpImg.transpose((2, 0, 1))
    return tmpImg

def load_processed_image(file_path):
    image = cv2.imread(file_path)
    tmpImg = preprocess(image)
    return tmpImg

def load_multi_image(data_paths, target_dtype: mm.DataType = mm.DataType.FLOAT32) -> np.ndarray:
    # Load multiple pre-processed image into a NCHW style ndarray
    images = []
    with open(data_paths[0]) as f:
        for line in f.readlines():
            line = line.strip('\n')
            line = MSRA_B_DATASETS_PATH + os.sep + line
            images.append(load_processed_image(line)[np.newaxis, :])
    ret = np.concatenate(tuple(images))
    return np.ascontiguousarray(ret.astype(dtype=get_numpy_dtype_by_datatype(target_dtype)))

class FixedCalibData(mm.CalibDataInterface):
    def __init__(self, shape: mm.Dims, data_type: mm.DataType, max_samples: int,
                 data_paths):
        super().__init__()
        self.shape_ = shape
        self.data_type_ = data_type
        self.batch_size_ = shape.GetDimValue(0)
        self.max_samples_ = min(max_samples, len(data_paths))
        self.data_paths_ = data_paths
        self.current_sample_ = None
        self.outputed_sample_count = 0

    def get_shape(self):
        return self.shape_

    def get_data_type(self):
        return self.data_type_

    def get_sample(self):
        return self.current_sample_

    def next(self):
        beg_ind = self.outputed_sample_count
        end_ind = self.outputed_sample_count + self.max_samples_
        if end_ind > self.max_samples_:
            return mm.Status(mm.Code.OUT_OF_RANGE, "End reached")

        self.current_sample_ = load_multi_image(self.data_paths_, 
                                                target_dtype=self.data_type_)
        self.outputed_sample_count = end_ind
        return mm.Status.OK()

    def reset(self):
        self.current_sample_ = None
        self.outputed_sample_count = 0
        return mm.Status.OK()

def calibrate(args, network : mm.Network, config : mm.BuilderConfig):
    # 创建量化工具并设置量化统计算法
    calib_data = FixedCalibData(shape = mm.Dims([args.batch_size, 3, args.input_height, args.input_width]),
                                data_type = DataType.FLOAT32,
                                max_samples = 10,
                                data_paths = [args.file_list])
    calibrator = mm.Calibrator([calib_data])
    assert calibrator is not None
    # 设置量化统计算法，支持线性统计算法（LINEAR_ALGORITHM）及加强的最小化量化噪声算法（EQM_ALGORITHM）。
    assert calibrator.set_quantization_algorithm(mm.QuantizationAlgorithm.LINEAR_ALGORITHM).ok()
    # 打开设备
    with mm.System() as mm_sys:
        dev_count = mm_sys.device_count()
        print("Device count: ", dev_count)
        if args.device_id >= dev_count:
            print("Invalid device set!")
            abort()
        # 打开MLU设备
        dev = mm.Device()
        dev.id = args.device_id
        assert dev.active().ok()
    # 进行量化
    assert calibrator.calibrate(network, config).ok()

def main():
    args = argparse.ArgumentParser()
    args.add_argument('--pt_model', dest = 'pt_model', default = '/workspace/saved_pt/2d_unet_0.pt',
            required = True, type = str, help = 'tf output graph')
    args.add_argument('--output_model', dest = 'output_model', default = '/workspace/offline_models/2d_unet_0.model',
            type = str, help = 'output model path')
    args.add_argument('--precision', dest = 'precision', default = 'qint8_mixed_float16',
            type = str, help = 'precision, qint8_mixed_float16 qint8_mixed_float32 force_float16 force float32 qint16_mixed_float32 are supported')
    args.add_argument('--file_list', dest = 'file_list', default = 'file_list',
            type = str, help = 'image list file path, file contains input image paths for calibration')
    args.add_argument('--batch_size', dest = 'batch_size', default = 1,
            type = int, help = 'batch_size')
    args.add_argument('--input_width', dest = 'input_width', default = 320,
            type = int, help = 'model input width')
    args.add_argument('--input_height', dest = 'input_height', default = 320,
            type = int, help = 'model input height')
    args.add_argument('--device_id', dest = 'device_id', default = 0,
            type = int, help = 'mlu device id, used for calibration')
    args = args.parse_args()

    supported_precision = ['qint8_mixed_float16', 'qint8_mixed_float32', 'force_float16', 'force_float32', 'qint16_mixed_float32']
    if args.precision not in supported_precision:
        print('precision [' + args.precision + ']', 'not supported')
        exit()

    network = torch_parser(args)
    config = generate_model_config(args)
    if args.precision.find('qint') != -1:
        print('do calibrate...')
        calibrate(args, network, config)
    print('build model...')
    # 生成模型
    builder = mm.Builder()
    model = builder.build_model('magicmind model', network, config)
    assert model is not None
    # 将模型序列化为离线文件
    assert model.serialize_to_file(args.output_model).ok()
    print("Generate model done, model save to %s" % args.output_model)

if __name__ == "__main__":
    main()



import argparse
import distutils.util


def get_argparser():
    """Parse commond line parameters.

    Returns:
        - Returns a argparse.Namespace with all command line parameters.
    """
    parent_parser = argparse.ArgumentParser(add_help=False)
    # parameters about network configuration
    net_config = parent_parser.add_argument_group(
        title="net_config", description="parameters for network configuration"
    )
    net_config.add_argument(
        "--input_dims",
        dest="input_dims",
        type=int,
        nargs="+",
        action="append",
        help="Input shapes by order. -1 represents dynamic axis. Specify dims for "
        + "for each input by --input_dims 1 3 224 224 --input_dims 1 3 24 24 --input_dims "
        + "... if there are multiple inputs.",
    )
    net_config.add_argument(
        "--batch_size",
        dest="batch_size",
        type=int,
        nargs="+",
        help="Input batchsize by order, will override all highest"
        + "dimensions for all inputs and not affect unrank/scalar inputs.",
    )
    net_config.add_argument(
        "--rgb2bgr",
        dest="rgb2bgr",
        default=False,
        type=lambda x: bool(distutils.util.strtobool(x)),
        help="convert RGB to BGR for first layer's Conv/BatchNorm/Scale of network",
    )

    # parameters about model builder
    model_builder = parent_parser.add_argument_group(
        title="model_builder", description="parameters for model building"
    )
    model_builder.add_argument(
        "--mlu_arch",
        dest="mlu_arch",
        type=str,
        nargs="+",
        default="mtp_372",
        choices=["mtp_220", "mtp_270", "mtp_290", "mtp_372", "tp_322", "mtp_592"],
        help="Target arch for mlu dev. Unset means all.",
    )
    model_builder.add_argument(
        "--cluster_num",
        dest="cluster_num",
        type=int,
        nargs="+",
        action="append",
        help="Allow users to flexibly specify the cluster.",
    )
    model_builder.add_argument(
        "--precision",
        dest="precision",
        type=str,
        default="force_float32",
        choices=[
            "force_float32",
            "force_float16",
            "qint16_mixed_float16",
            "qint8_mixed_float16",
            "qint16_mixed_float32",
            "qint8_mixed_float32",
        ],
        help="Mix precision mode.",
    )
    model_builder.add_argument(
        "--means",
        dest="means",
        type=float,
        nargs="+",
        action="append",
        help="Means for inputs by order. MUST input with vars.",
    )
    model_builder.add_argument(
        "--vars",
        dest="vars",
        type=float,
        nargs="+",
        action="append",
        help="Vars for inputs by order. MUST input with means.",
    )
    model_builder.add_argument(
        "--toolchain_path",
        dest="toolchain_path",
        type=str,
        default=None,
        help="Cross compile toolchain path for tp_322 and mtp_220.",
    )
    model_builder.add_argument(
        "--input_layout",
        dest="input_layout",
        type=str,
        nargs="+",
        choices=["NCHW", "NHWC", "NCT", "NTC", "NCDHW", "NDHWC"],
        help="Convert input layouts from channel last to channel second"
        + "(or the opposite) by order.",
    )
    model_builder.add_argument(
        "--output_layout",
        dest="output_layout",
        type=str,
        nargs="+",
        choices=["NCHW", "NHWC", "NCT", "NTC", "NCDHW", "NDHWC"],
        help="Output data types by order.",
    )
    model_builder.add_argument(
        "--dim_range_min",
        dest="dim_range_min",
        type=int,
        nargs="+",
        action="append",
        help="lower dimension boundary for each input. specify like "
        "--dim_range_min input0_shape[0] input0_shape[1] input0_shape[2] input0_shape[3]"
        "--dim_range_min input1_shape[0] input1_shape[1] input1_shape[2] input1_shape[4]"
        "... --dim_range_min inputn_shape[0] inputn_shape[2]...",
    )
    model_builder.add_argument(
        "--dim_range_max",
        dest="dim_range_max",
        type=int,
        nargs="+",
        action="append",
        help="upper dimension boundary for each input. specify like "
        "--dim_range_max input0_shape[0] input0_shape[1] input0_shape[2] input0_shape[3]"
        "--dim_range_max input1_shape[0] input1_shape[1] input1_shape[2] input1_shape[4]"
        "... --dim_range_max inputn_shape[0] inputn_shape[2]...",
    )
    model_builder.add_argument(
        "--build_config",
        dest="build_config",
        type=str,
        default=None,
        help="Additional json build config for build. The param in config json will"
        + "override the same setted arg params.",
    )

    # cambricon-note: add dynamic_shape
    model_builder.add_argument(
        "--dynamic_shape",
        dest="dynamic_shape",
        type=str,
        default="true",
        choices=["true", "false"],
        help="used to control graph_shape_mutable is true or false",
    )
    model_builder.add_argument(
        "--computation_preference",
        dest="computation_preference",
        type=str,
        default="auto",
        choices=["auto", "fast", "high_precision"],
        help="Specifies the computation mode of operations in network.",
    )
    model_builder.add_argument(
        "--type64to32_conversion",
        dest="type64to32_conversion",
        type=str,
        default="false",
        choices=["false", "true"],
        help="Specifies whether to enable the data conversion from 64-bit to 32-bit. This option should be enabled when parsing a framework model containing bit data (not supported by MagicMind) with IParser. This configuration is also valid for the quantized calibration. If there is a risk of accuracy overflow during this conversion, MagicMind will emit warning messages.",
    )
    model_builder.add_argument(
        "--conv_scale_fold",
        dest="conv_scale_fold",
        type=str,
        default="false",
        choices=["false", "true"],
        help="Whether to prepend the multiplication and addition operations to the convolution weights after the convolution operation. This configuration is also valid for the quantization calibration. If you want to use this optimization for quantization, you need to set this option in in advance during the quantization calibration phase to ensure that the quantization parameters during the calibration process are consistent with the actual data distribution. Note that this optimization may have an impact on the range of the weight distribution and turning on this optimization in lower bit-width uniform quantization algorithms may cause the accuracy not to meet the requirements.",
    )

    # parameters about model calibration
    calibration = parent_parser.add_argument_group(
        title="model_calibration", description="parameters for model calibration"
    )
    calibration.add_argument(
        "--device_id",
        dest="device_id",
        type=int,
        default=0,
        help="MLU device id, used for calibration.",
    )
    calibration.add_argument(
        "--random_calib_range",
        dest="random_calib_range",
        type=float,
        nargs="*",
        help="Set random range for calibration. Will override path and filelist.",
    )
    calibration.add_argument(
        "--file_list",
        dest="file_list",
        type=str,
        nargs="+",
        help="Input file list path by order. For calibration only. MUST "
        + "input with calibration_data_path.",
    )
    calibration.add_argument(
        "--calibration_data_path",
        dest="calibration_data_path",
        type=str,
        default=None,
        help="Directory for calibration data. MUST input with file_list.",
    )
    calibration.add_argument(
        "--calibration",
        dest="calibration",
        type=lambda x: bool(distutils.util.strtobool(x)),
        default="false",
        help="To do calibration or not. Will use range [-1,1] and skip "
        + "calibration if no file or range is set.",
    )
    calibration.add_argument(
        "--rpc_server",
        dest="rpc_server",
        type=str,
        default=None,
        help="Set remote address for calibration.",
    )
    calibration.add_argument(
        "--calibration_algo",
        dest="calibration_algo",
        type=str,
        default="linear",
        choices=["linear", "eqnm"],
        help="Set quantization algorithm for calibration.",
    )
    calibration.add_argument(
        "--weight_quant_granularity",
        dest="weight_quant_granularity",
        type=str,
        default="per_tensor",
        choices=["per_tensor", "per_axis"],
        help="quantization granularity for weights.",
    )
    calibration.add_argument(
        "--activation_quant_algo",
        dest="activation_quant_algo",
        type=str,
        default="symmetric",
        choices=["symmetric", "asymmetric"],
        help="specify symmetric quantization or asymmetric quantization.",
    )

    # parameters about model building and serialization
    build_and_serilaize = parent_parser.add_argument_group(
        title="model_build_and_serialize",
        description="parameters for model building and serialization",
    )
    build_and_serilaize.add_argument(
        "--input_dtypes",
        dest="input_dtypes",
        type=str,
        choices=[
            "INT8",
            "INT16",
            "INT32",
            "UINT8",
            "UINT16",
            "UINT32",
            "HALF",
            "FLOAT",
            "BOOL",
            "QINT8",
            "QINT16",
        ],
        nargs="+",
        help="Input data types by order for inference (will not affect"
        + "calibration).",
    )
    build_and_serilaize.add_argument(
        "--output_dtypes",
        dest="output_dtypes",
        type=str,
        choices=[
            "INT8",
            "INT16",
            "INT32",
            "UINT8",
            "UINT16",
            "UINT32",
            "HALF",
            "FLOAT",
            "BOOL",
            "QINT8",
            "QINT16",
        ],
        nargs="+",
        help="Output data types by order.",
    )
    build_and_serilaize.add_argument(
        "--magicmind_model",
        dest="magicmind_model",
        type=str,
        default="./model",
        help="File path for output serialization model file.",
    )

    parent_parser.add_argument(
        "--plugin_path",
        dest="plugin_path",
        type=str,
        nargs="*",
        help="Plugin library paths to link with.",
    )

    # parameters about model calibration
    framework_parser = parent_parser.add_argument_group(
        title="framework_parser", description="parameters for pytorch/tf/caffe/onnx parser"
    )

    framework_parser.add_argument(
        "--tf_pb",
        dest="tf_pb",
        type=str,
        default=None,
        help="TensorFlow pb file for TensorFlow parser.",
    )
    framework_parser.add_argument(
        "--input_names",
        dest="input_names",
        type=str,
        nargs="+",
        default=None,
        help="Input names for TensorFlow parser.",
    )
    framework_parser.add_argument(
        "--output_names",
        dest="output_names",
        type=str,
        nargs="+",
        default=None,
        help="Output names for TensorFlow parser.",
    )
    # pytorch
    framework_parser.add_argument(
        "--pytorch_pt",
        dest="pytorch_pt",
        type=str,
        default=None,
        help="PyTorch pt file path for PyTorch parser.",
    )
    framework_parser.add_argument(
        "--pt_input_dtypes",
        dest="pt_input_dtypes",
        type=str,
        nargs="+",
        choices=[
            "INT8",
            "INT16",
            "INT32",
            "UINT8",
            "UINT16",
            "UINT32",
            "HALF",
            "FLOAT",
            "BOOL",
            "QINT8",
            "QINT16",
        ],
        default="FLOAT",
        help="Input data types by order for parsing PyTorch pt",
    )
    # onnx
    framework_parser.add_argument(
        "--onnx",
        dest="onnx",
        type=str,
        default=None,
        help="Onnx file path for ONNX parser.",
    )
    # caffe
    framework_parser.add_argument(
        "--prototxt",
        dest="prototxt",
        type=str,
        default=None,
        help="prototxt file path for Caffe parser.",
    )
    framework_parser.add_argument(
        "--caffemodel",
        dest="caffemodel",
        type=str,
        default=None,
        help="Caffemodel file path for Caffe parser.",
    )
    return parent_parser



import cv2
import numpy as np

def imrescale(img, scale):
    h, w = img.shape[:2]
    max_long_edge = max(scale)
    max_short_edge = min(scale)
    scale_factor = min(max_long_edge / max(h, w),
                        max_short_edge / min(h, w))
    new_size = int(w * float(scale_factor) + 0.5), int(h * float(scale_factor) + 0.5)
    resized_img = cv2.resize(
            img, new_size, dst=None, interpolation=cv2.INTER_LINEAR)
    new_h, new_w = resized_img.shape[:2]
    w_scale = new_w / w
    h_scale = new_h / h
    scale_factor = np.array([w_scale, h_scale, w_scale, h_scale])
    return resized_img, scale_factor

def imnormalize(img, mean, std, to_rgb=True):
    """Inplace normalize an image with mean and std.
    Args:
        img (ndarray): Image to be normalized.
        mean (ndarray): The mean to be used for normalize.
        std (ndarray): The std to be used for normalize.
        to_rgb (bool): Whether to convert to rgb.
    Returns:
        ndarray: The normalized image.
    """
    # cv2 inplace normalization does not accept uint8
    img = img.astype(np.float32)
    mean = np.float64(mean.reshape(1, -1))
    stdinv = 1 / np.float64(std.reshape(1, -1))
    if to_rgb:
        cv2.cvtColor(img, cv2.COLOR_BGR2RGB, img)  # inplace
    cv2.subtract(img, mean, img)  # inplace
    cv2.multiply(img, stdinv, img)  # inplace
    img = img.transpose(2,0,1)
    return np.expand_dims(img, axis=0)

def preprocess(image):
    scale = (2048, 1024)
    mean = np.array([123.675, 116.28 , 103.53])
    std = np.array([58.395, 57.12 , 57.375])
    resized_img, scale_size = imrescale(image, scale)
    img = imnormalize(resized_img, mean, std)
    return img, scale_size



import argparse
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval
from utils import Record


def main():
    args = argparse.ArgumentParser(description='Evaluate')
    args.add_argument('--ann_file', dest = 'ann_file', required = True, type = str, help = 'annotation file path')
    args.add_argument('--res_file', dest = 'res_file', required = True, type = str, help = 'result file path')
    args.add_argument('--res2_file', dest = 'res2_file', required = False, type = str, help = 'result file path')
    args.add_argument('--output_file', dest = 'output_file', required = False, type = str, help = 'save result to file')
    args.add_argument('--iou_type', dest = 'iou_type', required = False, default = "keypoints", type = str, help = 'save result to file')
    args = args.parse_args()

    coco_gt = COCO(args.ann_file)
    coco_dt = coco_gt.loadRes(args.res_file)
    iou_type = args.iou_type
    coco_eval = COCOeval(coco_gt, coco_dt, iou_type)
    coco_eval.params.imgIds = coco_gt.getImgIds()
    coco_eval.evaluate()
    coco_eval.accumulate()
    coco_eval.summarize()
    stat1 = coco_eval.stats[1]
    stat2 = 0.0;
    if args.res2_file is not None:
        coco_gt = COCO(args.ann_file)
        coco_dt = coco_gt.loadRes(args.res2_file)
        coco_eval = COCOeval(coco_gt, coco_dt, iou_type)
        coco_eval.params.imgIds = coco_gt.getImgIds()
        coco_eval.evaluate()
        coco_eval.accumulate()
        coco_eval.summarize()
        stat2 = coco_eval.stats[1]
        
    if args.output_file is not None:
        output_file = Record(args.output_file)
        output_file.write("AP IoU=0.50:%f"%(stat1), False);
        if args.res2_file is not None:
            output_file.write("AP IoU=0.50:%f"%(stat2), False);

if __name__ == "__main__":
    main()

